/*! ********************************************
 * MidiController.hpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_MidiController_h
#define INCLUDE_MidiController_h

#ifndef INCLUDE_GrayBox_h
#include "GrayBox.hpp"
#endif

namespace sora {
namespace graybox {
namespace util {
class MidiController : public IGrayBoxModule
{
protected:
	MidiEvent* _eventBuffer = nullptr;
	int32 _eventBufferSize = 64;
	int32 _usedEvents = 0;
public:
	int32 midiOutPorts = 0xFFFF;
// Method
	void clearMIDIEventBuffer();
	int32 getEventBufferSize() { return _eventBufferSize; }
	int32 setEventBufferSize(int32 size);
// MIDI Controlle
	void noteOff(const int32& ch, const int32& key, const int32& offsetFrames = 0);
	void noteOn(const int32& ch, const int32& key, const int32& velo, const int32& offsetFrames = 0);
	void keyAfterTouch(const int32& ch, const int32& key, const int32& val, const int32& offsetFrames = 0);
	void controlChange(const int32& ch, const int32& ccn, const int32& val, const int32& offsetFrames = 0);
	void programChange(const int32& ch, const int32& val, const int32& offsetFrames = 0);
	void channelAfterTouch(const int32& ch, const int32& val, const int32& offsetFrames = 0);
	void pitchBendChange(const int32& ch, const int32& lsb, const int32& msb, const int32& offsetFrames = 0);
	void pitchBendChangeW(const int32& ch, const int32& val, const int32& offsetFrames = 0);
	void systemExclusive(const uint8* msg, const int32& dataSize, const int32& offsetFrames = 0);
	void midiMessage(const int32& type, const int32& ch = 0, const int32& val0 = 0, const int32& val1 = 0, 
	                 const uint8* msg = nullptr, const int32& dataSize = 0, const int32& offsetFrames = 0);
	void midiMessage_P(const int32& portNum = 0xffff, const int32& type = 0x80, const int32& ch = 0, const int32& val0 = 0, const int32& val1 = 0,
	                 const uint8* msg = nullptr, const int32& dataSize = 0, const int32& offsetFrames = 0);
// Constructor
	MidiController();
	MidiController(int32 eventBufferSize);
// Copy Constructor
	MidiController(const MidiController&) = delete;
	MidiController& operator = (const MidiController&) = delete;
// Move Constructor
	MidiController(MidiController&&)noexcept;
	MidiController& operator = (MidiController&&)noexcept;
// Destructor
	~MidiController();
// Interface - IfGrayBoxModule -----
	void updateSampleRate(float) override {}
	void updateAudioChannels(int32) override {}
	void updateProcessBlockSize(int32) override {}
	
	void resetModule() override;
	void initiModule() override;
	void connectModule(GrayBox*) override;
	void disconnectModule() override;
	void updateBypassStatus(bool) override;
	
	void startProcess() override;
	void stopProcess() override;
	void suspendProcess() override;
	void resumeProcess() override;
	void process(ProcessData*) override;
};
}}}
#endif