/*! ********************************************
 * SynthesizerFM2.hpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SynthesizerFM2_h
#define INCLUDE_SynthesizerFM2_h

#ifndef INCLUDE_MultiTimberSynthBase_h
#include "MultiTimberSynthBase.hpp"
#endif

#ifndef INCLUDE_SampleData_h
#include "SampleData.hpp"
#endif

namespace sora {
namespace graybox {
namespace synth {
class TonePresetFM
{
public:
	char toneName[17]={"                "};	// ASCII 16Chars 
	int32 key = 64;        // 0 - 127     *Percussion
	int32 pan = 64;        // 0 - 127     *Percussion
	int32 lfoWave = 0;     // 0 - 3
	int32 lfoFreq = 2048;  // 0 - 16383   LFO - Phase Update Timer Parameter
	int32 fl = 127;        // 0 - 127    // Feed Back Level
	int32 strc = 0;        // 0 - 7	  // strcture
	// operator 0
	//int32 velCurve_0 = 0;
	int32 tlVelCtrl_0 = 0;// 0 - 7      Total Level Contorol - Velocity
	int32 mul_0 = 1;      // 0 - 15     Multiple       [0] = 1/2 Cycle, [1] = 1/1, [2] = 2/1, [3] = 3/1...
	int32 tune_0 = 8192;  // 0 - 16383  center 8192, octave = 8192
	int32 lms_0 = 0;      // 0 - 15      LFO - Modulation Depth Senstivity
	int32 wave_0 = 0;     // 0 - 7
	int32 tl_0 = 0;       // 0 - 127
	int32 ar_0 = 124;     // 0 - 127
	int32 dr_0 = 100;     // 0 - 127
	int32 sl_0 = 0;       // 0 - 127
	int32 rr_0 = 120;     // 0 - 127
	int32 kf_0 = 0;       // 0 - 3      Key Follow Sensitivity         *Instrument 
	int32 kfStart_0 = 48; // 0 - 127    Key Number - Start Key Follow  *Instrument
	// operator 1
	//int32 velCurve_1 = 0;
	int32 tlVelCtrl_1 = 0;// 0 - 7
	int32 mul_1 = 1;      // 0 - 15
	int32 tune_1 = 8192;  // 0 - 16383
	int32 lms_1 = 0;     // 0 - 15      LFO - Modulation Depth Senstivity
	int32 wave_1 = 0;     // 0 - 7
	int32 tl_1 = 0;       // 0 - 127
	int32 ar_1 = 124;     // 0 - 127
	int32 dr_1 = 100;     // 0 - 127
	int32 sl_1 = 0;       // 0 - 127
	int32 rr_1 = 120;     // 0 - 127
	int32 kf_1 = 0;       // 0 - 3
	int32 kfStart_1 = 48; // 0 - 127
	
	void copy(const TonePresetFM& preset);
	void reset();
};
class SynthesizerFM2 final : public MultiTimberSynthBase
{
public:
	static constexpr int32 WAVE_TABLE_SIZE_BITS = 11;                   // 16384 (DataEntry MSB LSB) / 8 (waves)
	static constexpr int32 WAVE_TABLE_SIZE = 1 << WAVE_TABLE_SIZE_BITS; // 16384 (DataEntry MSB LSB) / 8 (waves)
	static constexpr int32 PG_PHASE_FP = 16; // 16bit fixed-point  (precision 1 / 65536)
	static constexpr int32 OVER_TONES = 16;  // for Sinewave Synthesis.
private:
	class VoiceUnitFM2 final : public VoiceUnitBase
	{
	// Tables
	private:
		// Log2 Volume Table
		float* _LOG2V_TABLE = nullptr;
		int32* _LOG2V_TABLE_I = nullptr;
		// Pow2 Table
		static constexpr int32 _POW2T_OCTV_BITS = SynthData::POW2T_OCTV_BITS;
		static constexpr int32 _POW2T_OCTV_MASK = SynthData::POW2T_OCTV_MASK;
		int32* _POW2_TABLE = nullptr;
		int32* _POW2T_RS_TABLE = nullptr;
		int32* _POW2T_KEY_TABLE = nullptr;
		// EG
		static constexpr int32 _EG_PRAM_OCTV_LS = SynthData::POW2T_OCTV_BITS - 3;
		static constexpr int32 _EG_CNT_FP = SynthData::EG_CNT_FP;
		static constexpr int32 _EG_MAX_CNT = SynthData::EG_MAX_CNT;
		static constexpr int32 _EG_CNT_ENV_ATK = SynthData::EG_CNT_ENV_ATK;
		static constexpr int32 _EG_CNT_ENV_BTM = SynthData::EG_CNT_ENV_BTM;
		static constexpr int32 _EG_SUSTAIN_BTM = SynthData::EG_SUSTAIN_BTM;
		int32* _EG_TOTAL_LEVEL_TABLE = nullptr;
		int32* _EG_LEVEL_TABLE_UP = nullptr;
		int32* _EG_LEVEL_TABLE_DW = nullptr;
		int32* _EG_KEY_TABLE = nullptr;
		int32* _EG_POW2T_RS_TABLE = nullptr;
		int32* _EG_CNT_GATE_TABLE = nullptr;
		// PG
		static constexpr int32 _WAVE_TABLE_SIZE_BITS = SynthesizerFM2::WAVE_TABLE_SIZE_BITS;
		static constexpr int32 _WAVE_TABLE_MASK  = SynthesizerFM2::WAVE_TABLE_SIZE - 1;
		static constexpr int32 _PG_PHASE_FP      = SynthesizerFM2::PG_PHASE_FP;
		static constexpr int32 _PG_PHASE_FP_MASK = (1 << _PG_PHASE_FP) - 1;
		static constexpr int32 _PG_PHASE_MASK    = (SynthesizerFM2::WAVE_TABLE_SIZE << _PG_PHASE_FP) - 1;
		// FM
		int32* _MUL_TABLE = nullptr;
		// LFO
		static constexpr int32 _LFO_WAVE_TABLE_SIZE_BITS = SynthData::LFO_WAVE_TABLE_SIZE_BITS;
		static constexpr int32 _LFO_TIMER_INIT_VALUE = SynthData::LFO_TIMER_INIT_VALUE;
		static constexpr int32 _LFO_WAVE_TABLE_MASK  = SynthData::LFO_WAVE_TABLE_SIZE - 1;
		int32* _LFO_WAVE_TABLES = nullptr;
	// Synthesizer Data
	public:
		VoicePart* VOICE_PARTS = nullptr;
		int32* WAVE_TABLES = nullptr;
		TonePresetFM* TONE_PRESETS = nullptr;
		SampleData* SAMPLE_DATAS = nullptr;
		SamplerPreset* SAMPLER_PRESETS = nullptr;
		int32 pgSamplePow2tIdx = 0;  // PG PhaseStep (Pow2Table Index Offset) - SamplingRate
		int32 pgKeyPow2tIdx    = 0;  // PG PhaseStep (Pow2Table Index Offset) - ReferenceKey (NoteNumber 0)
		int32 egPow2tIdx       = 0;  // EG PhaseStep (Pow2Table Index Offset)
		int32 egPhaseStepSndOff= 0;
		int32 lfoTimerCoef     = 0;  // Coefficient Value - LFO Phase Update Timer
	// VoiceUnit Status
		// Pitch
	private:	
		int32 _ptBend     = 0;   // pitchBend
		int32 _ptBendSens = 0;   // pitchBend Sensitivity
		// LFO
		int32 _lfoPhase     = 0;  // LFO Phase
		int32 _lfoTimer     = 0;  // LFO Phase Update Timer
		int32 _lfoTimerStep = 0;  // Update Timer Step
		int32 _lfoMod_0      = 0;  // Modulation Depth
		int32 _lfoMod_1      = 0;  // Modulation Depth
		int32 _lfoModSens_0  = 0;  // Modulation Sensitivity
		int32 _lfoModSens_1  = 0;  // Modulation Sensitivity
		int32 _lfoOutPM_0    = 0;  // output - phase modulation
		int32 _lfoOutPM_1    = 0;  // output - phase modulation
		int32 _lfoOutAM_0    = 0;  // output - amplitude modulation
		int32 _lfoOutAM_1    = 0;  // output - amplitude modulation
	// Operator 0,1 ---
		int32* _lfoWaveTable = nullptr;
		int32 _pgBend = 0;   // pitchBend
		int32 _fdbk   = 0;   // pitchBend Sensitivity
	// Operator 0 ---
		// Wave
		int32* _waveTable_0  = nullptr;
		int32 _pgKey_0       = 0;
		int32 _pgPhase_0     = 0;
		int32 _pgPhaseStep_0 = 0;
		// EG
		/*
		- Envelop Generator State -
		[ 5674 3210 ] 8Bits Flag Status (0 = deActive)
		[ ---- RSDA ] "-" = Reserved Bits
			A = Attack  stage
			D = Decay   stage
			S = Sustain stage
			R = Release stage
		*/
		int32 _egState_0      = 0;
		int32* _egLevelTable_0= nullptr;
		int32 _egCount_0      = 0;
		int32 _egCountStep_0  = 0;
		//int32 _egCountStepAR= 0;
		int32 _egCountStepDR_0= 0;
		int32 _egCountStepRR_0= 0;
		int32 _egCntEnvSus_0  = 0;
		int32 _egCntEnvBtm_0  = 0;
		int32 _egGate_0       = 0;
		int32 _egSL_0         = 0;
		int32 _egTL_0         = 0;
		int32 _egTLv_0        = 0;
	// Operator 1 ---
		// Wave
		int32* _waveTable_1  = nullptr;
		int32 _pgKey_1       = 0;
		int32 _pgPhase_1     = 0;
		int32 _pgPhaseStep_1 = 0;
		// EG
		int32 _egState_1      = 0;
		int32* _egLevelTable_1= nullptr;
		int32 _egCount_1      = 0;
		int32 _egCountStep_1  = 0;
		//int32 _egCountStepAR= 0;
		int32 _egCountStepDR_1= 0;
		int32 _egCountStepRR_1= 0;
		int32 _egCntEnvSus_1  = 0;
		int32 _egCntEnvBtm_1  = 0;
		int32 _egGate_1       = 0;
		int32 _egSL_1         = 0;
		int32 _egTL_1         = 0;
		int32 _egTLv_1        = 0;
	// Sampler
		int32 _pgPhaseAFP = 0;  // PG Phase (After the Fixed Point)
		int32 _plbkE      = 0;  // PlayBack End Frame   [plbkE <= channelPerSamples]
		int32 _loopE      = 0;  // Loop End Frame       [loopE <= plbkE]
		int32 _loopF      = 0;  // Loop Frames          [val = loopE - lpStart]
		int32 _loopFk     = 0;  // Loop Frames          * after -> keyOff
	
	public:
		void keyOff() override;
		void soundOff() override;
	private:
		int32 _shiftEGState_0();
		int32 _shiftEGState_1();
	public:
		void keyOn(const int32&, const int32&, const int32&) override;
		void keyOnSampler(const int32&, const int32&, const int32&);
		void updPartBendSens(const int32&) override;
		void updPartPitchBend(const int32&) override;
		void updPartModulation(const int32&) override;
		void updateLfoFrequency(const int32&);
	public:
		void setup(SynthData*);
	// process ---
	protected:
		void _proc0(float* const&, float* const&, const int32&) override;
		void _proc1(float* const&, float* const&, const int32&) override;
		void _proc2(float* const&, float* const&, const int32&) override;
		void _proc3(float* const&, float* const&, const int32&) override;
		void _proc4(float* const&, float* const&, const int32&) override;
		void _proc5(float* const&, float* const&, const int32&) override;
		void _proc6(float* const&, float* const&, const int32&) override;
		void _proc7(float* const&, float* const&, const int32&) override;
		void _proc8(float* const&, float* const&, const int32&) override;
		void _proc9(float* const&, float* const&, const int32&) override;
	};

// Factory -----
private:	
	void _recreateVU(int32 maxPoly) override;
	void _resetKeyParams() override;
// Synthesizer - Data -----
private:
	SynthData* _prvSynthData = nullptr;
	SynthData* _synthData = nullptr;
	int32* waveTables = nullptr;  // 128 * 128
	SampleData* sampleDatas = nullptr;
public:
	TonePresetFM* tonePresets = nullptr;  // [0 - 127](ProgramChange), [128 - 255](Percussion Key)
	SamplerPreset* samplerPresets = nullptr;
	
	void getWaveForm(int32 id, float* buffer, int32 bufferSize);
	void setWaveForm(int32 id, const float* buffer, int32 bufferSize);
	SampleData* getSampleData(int32 id){ return &sampleDatas[id]; }
	void setSampleData(int32 id, SampleData&& sampleData);
	void setSampleData(int32 id, const SampleData& sampleData, bool deepCopy = true);
	void loadPCMSamples(int32 id, float* wave, int32 samples, int32 ch);
	void loadPCMData(int32 id, uint8* dat, int32 sampleFrames, int32 ch, int32 bitRate, bool sign, bool bigEndian);
	void clearSampleData(int32 id);
	void factoryDataReset() override;
// Synthesizer - main -----
private:
	VoiceUnitFM2* _voiceUnitExt = nullptr;
// Constructor
public:
	SynthesizerFM2() : SynthesizerFM2(32, 440.0f, 1.0f, true){}
	SynthesizerFM2(int32 maxPoly = 32, float refPitch = 440.0f, float masVol = 1.0f, bool privateTableData = true);// Controls ---
// Copy Constructor
	SynthesizerFM2(const SynthesizerFM2&) = delete;
	SynthesizerFM2& operator = (const SynthesizerFM2&) = delete;
// Move Constructor
	SynthesizerFM2(SynthesizerFM2&&)noexcept;
	SynthesizerFM2& operator = (SynthesizerFM2&&)noexcept;
// Destructor
	~SynthesizerFM2();
// MidiEvent ---
private:	
	void _noteOn(const MidiEvent&) override;
	void _readSysExData(const MidiEvent&) override;
};
}}}
#endif