/*! ********************************************
 * SynthesizerFM4.hpp
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
#ifndef INCLUDE_SynthesizerFM4_h
#define INCLUDE_SynthesizerFM4_h

#ifndef INCLUDE_MultiTimberSynthBase_h
#include "MultiTimberSynthBase.hpp"
#endif

#ifndef INCLUDE_SampleData_h
#include "SampleData.hpp"
#endif

namespace sora {
namespace graybox {
namespace synth {
class TonePresetFM4
{
public:
	char toneName[17]={"                "};	// ASCII 16Chars 
	int32 key = 64;        // 0 - 127     *Percussion
	int32 pan = 64;        // 0 - 127     *Percussion
	int32 lfoWave = 0;     // 0 - 3
	int32 lfoFreq = 2048;  // 0 - 16383   LFO - Phase Update Timer Parameter
	int32 fl = 127;        // 0 - 127    // Feed Back Level
	int32 strc = 0;        // 0 - 7	  // strcture

	int32 tlVelCtrl[4] = { 0, 0, 0, 0 };// 0 - 7      Total Level Contorol - Velocity
	int32 mul[4] = { 1, 1, 1, 1 };      // 0 - 7      Multiple       [0] = 1/2 Cycle, [1] = 1/1, [2] = 2/1, [3] = 3/1...
	int32 tune[4] = { 8192, 8192, 8192, 8192 };  // 0 - 16383  center 8192, octave = 8192
	int32 lms[4] = { 0, 0, 0, 0 };      // 0 - 15      LFO - Modulation Depth Senstivity
	int32 wave[4] = { 0, 0, 0, 0 };     // 0 - 7
	int32 tl[4] = { 0, 0, 0, 0 };       // 0 - 127
	int32 ar[4] = { 124, 124, 124, 124 };     // 0 - 127
	int32 dr[4] = { 100, 100, 100, 100 };     // 0 - 127
	int32 sl[4] = { 0, 0, 0, 0 };       // 0 - 127
	int32 rr[4] = { 120, 120, 120, 120 };     // 0 - 127
	int32 kf[4] = { 0, 0, 0, 0 };       // 0 - 3      Key Follow Sensitivity         *Instrument 
	int32 kfStart[4] = { 48, 48, 48, 48 }; // 0 - 127    Key Number - Start Key Follow  *Instrument
	// operator 0
	int32& tlVelCtrl_0 = tlVelCtrl[0];
	int32& mul_0 = mul[0];
	int32& tune_0 = tune[0];
	int32& lms_0 = lms[0];
	int32& wave_0 = wave[0];
	int32& tl_0 = tl[0];
	int32& ar_0 = ar[0];
	int32& dr_0 = dr[0];
	int32& sl_0 = sl[0];
	int32& rr_0 = rr[0];
	int32& kf_0 = kf[0];
	int32& kfStart_0 = kfStart[0];
	// operator 1
	int32& tlVelCtrl_1 = tlVelCtrl[1];
	int32& mul_1 = mul[1];
	int32& tune_1 = tune[1];
	int32& lms_1 = lms[1];
	int32& wave_1 = wave[1];
	int32& tl_1 = tl[1];
	int32& ar_1 = ar[1];
	int32& dr_1 = dr[1];
	int32& sl_1 = sl[1];
	int32& rr_1 = rr[1];
	int32& kf_1 = kf[1];
	int32& kfStart_1 = kfStart[1];
	// operator 2
	int32& tlVelCtrl_2 = tlVelCtrl[2];
	int32& mul_2 = mul[2];
	int32& tune_2 = tune[2];
	int32& lms_2 = lms[2];
	int32& wave_2 = wave[2];
	int32& tl_2 = tl[2];
	int32& ar_2 = ar[2];
	int32& dr_2 = dr[2];
	int32& sl_2 = sl[2];
	int32& rr_2 = rr[2];
	int32& kf_2 = kf[2];
	int32& kfStart_2 = kfStart[2];
	// operator 3
	int32& tlVelCtrl_3 = tlVelCtrl[3];
	int32& mul_3 = mul[3];
	int32& tune_3 = tune[3];
	int32& lms_3 = lms[3];
	int32& wave_3 = wave[3];
	int32& tl_3 = tl[3];
	int32& ar_3 = ar[3];
	int32& dr_3 = dr[3];
	int32& sl_3 = sl[3];
	int32& rr_3 = rr[3];
	int32& kf_3 = kf[3];
	int32& kfStart_3 = kfStart[3];
	void copy(const TonePresetFM4& preset);
	void reset();
};
class SynthesizerFM4 final : public MultiTimberSynthBase
{
public:
	static constexpr int32 WAVE_TABLE_SIZE_BITS = 11;                   // 16384 (DataEntry MSB LSB) / 8 (waves)
	static constexpr int32 WAVE_TABLE_SIZE = 1 << WAVE_TABLE_SIZE_BITS; // 16384 (DataEntry MSB LSB) / 8 (waves)
	static constexpr int32 PG_PHASE_FP = 16; // 16bit fixed-point  (precision 1 / 65536)
	static constexpr int32 OVER_TONES = 16;  // for Sinewave Synthesis.
private:
	class VoiceUnitFM4 final : public VoiceUnitBase
	{
	// Tables
	private:
		// Log2 Volume Table
		float* _LOG2V_TABLE = nullptr;
		int32* _LOG2V_TABLE_I = nullptr;
		// Pow2 Table
		static constexpr int32 _POW2T_OCTV_BITS = SynthData::POW2T_OCTV_BITS;
		static constexpr int32 _POW2T_OCTV_MASK = SynthData::POW2T_OCTV_MASK;
		int32* _POW2_TABLE = nullptr;
		int32* _POW2T_RS_TABLE = nullptr;
		int32* _POW2T_KEY_TABLE = nullptr;
		// EG
		static constexpr int32 _EG_PRAM_OCTV_LS = SynthData::POW2T_OCTV_BITS - 3;
		static constexpr int32 _EG_CNT_FP = SynthData::EG_CNT_FP;
		static constexpr int32 _EG_MAX_CNT = SynthData::EG_MAX_CNT;
		static constexpr int32 _EG_CNT_ENV_ATK = SynthData::EG_CNT_ENV_ATK;
		static constexpr int32 _EG_CNT_ENV_BTM = SynthData::EG_CNT_ENV_BTM;
		static constexpr int32 _EG_SUSTAIN_BTM = SynthData::EG_SUSTAIN_BTM;
		int32* _EG_TOTAL_LEVEL_TABLE = nullptr;
		int32* _EG_LEVEL_TABLE_UP = nullptr;
		int32* _EG_LEVEL_TABLE_DW = nullptr;
		int32* _EG_KEY_TABLE = nullptr;
		int32* _EG_POW2T_RS_TABLE = nullptr;
		int32* _EG_CNT_GATE_TABLE = nullptr;
		// PG
		static constexpr int32 _WAVE_TABLE_SIZE_BITS = SynthesizerFM4::WAVE_TABLE_SIZE_BITS;
		static constexpr int32 _WAVE_TABLE_MASK  = SynthesizerFM4::WAVE_TABLE_SIZE - 1;
		static constexpr int32 _PG_PHASE_FP      = SynthesizerFM4::PG_PHASE_FP;
		static constexpr int32 _PG_PHASE_FP_MASK = (1 << _PG_PHASE_FP) - 1;
		static constexpr int32 _PG_PHASE_MASK    = (SynthesizerFM4::WAVE_TABLE_SIZE << _PG_PHASE_FP) - 1;
		// FM
		int32* _MUL_TABLE = nullptr;
		// LFO
		static constexpr int32 _LFO_WAVE_TABLE_SIZE_BITS = SynthData::LFO_WAVE_TABLE_SIZE_BITS;
		static constexpr int32 _LFO_TIMER_INIT_VALUE = SynthData::LFO_TIMER_INIT_VALUE;
		static constexpr int32 _LFO_WAVE_TABLE_MASK  = SynthData::LFO_WAVE_TABLE_SIZE - 1;
		int32* _LFO_WAVE_TABLES = nullptr;
	// Synthesizer Data
	public:
		VoicePart* VOICE_PARTS = nullptr;
		int32* WAVE_TABLES = nullptr;
		TonePresetFM4* TONE_PRESETS = nullptr;
		SampleData* SAMPLE_DATAS = nullptr;
		SamplerPreset* SAMPLER_PRESETS = nullptr;
		int32 pgSamplePow2tIdx = 0;  // PG PhaseStep (Pow2Table Index Offset) - SamplingRate
		int32 pgKeyPow2tIdx    = 0;  // PG PhaseStep (Pow2Table Index Offset) - ReferenceKey (NoteNumber 0)
		int32 egPow2tIdx       = 0;  // EG PhaseStep (Pow2Table Index Offset)
		int32 egPhaseStepSndOff= 0;
		int32 lfoTimerCoef     = 0;  // Coefficient Value - LFO Phase Update Timer
	// VoiceUnit Status
	private:	
	// Sampler
		int32 _pgPhaseAFP = 0;  // PG Phase (After the Fixed Point)
		int32 _plbkE = 0;  // PlayBack End Frame   [plbkE <= channelPerSamples]
		int32 _loopE = 0;  // Loop End Frame       [loopE <= plbkE]
		int32 _loopF = 0;  // Loop Frames          [val = loopE - lpStart]
		int32 _loopFk = 0;  // Loop Frames          * after -> keyOff
	// Inst
		int32* _lfoWaveTable = nullptr;
		int32 _pgBend = 0;   // pitchBend
		int32 _fdbk = 0;   // pitchBend Sensitivity
		int32 _ptBend     = 0;   // pitchBend
		int32 _ptBendSens = 0;   // pitchBend Sensitivity
		// LFO
		int32 _lfoPhase = 0;  // LFO Phase
		int32 _lfoTimer = 0;  // LFO Phase Update Timer
		int32 _lfoTimerStep = 0;  // Update Timer Step
	// Operator ---
// LFO
		int32 _lfoMod[4] = { 0, 0, 0, 0 };  // Modulation Depth
		int32 _lfoModSens[4] = { 0, 0, 0, 0 };  // Modulation Sensitivity
		int32 _lfoOutPM[4] = { 0, 0, 0, 0 };  // output - phase modulation
		int32 _lfoOutAM[4] = { 0, 0, 0, 0 };  // output - amplitude modulation
		// Wave
		int32* _waveTable[4] = { nullptr, nullptr, nullptr, nullptr };
		int32 _pgKey[4] = { 0, 0, 0, 0 };
		int32 _pgPhase[4] = { 0, 0, 0, 0 };
		int32 _pgPhaseStep[4] = { 0, 0, 0, 0 };
		// EG
		/*
		- Envelop Generator State -
		[ 5674 3210 ] 8Bits Flag Status (0[4] =  deActive)
		[ ---- RSDA ] "-"[4] =  Reserved Bits
			A[4] =  Attack  stage
			D[4] =  Decay   stage
			S[4] =  Sustain stage
			R[4] =  Release stage
		*/
		int32 _egState[4] = { 0, 0, 0, 0 };
		int32* _egLevelTable[4] = { nullptr, nullptr, nullptr, nullptr };
		int32 _egCount[4] = { 0, 0, 0, 0 };
		int32 _egCountStep[4] = { 0, 0, 0, 0 };
		//int32 _egCountStepAR= 0;
		int32 _egCountStepDR[4] = { 0, 0, 0, 0 };
		int32 _egCountStepRR[4] = { 0, 0, 0, 0 };
		int32 _egCntEnvSus[4] = { 0, 0, 0, 0 };
		int32 _egCntEnvBtm[4] = { 0, 0, 0, 0 };
		int32 _egGate[4] = { 0, 0, 0, 0 };
		int32 _egSL[4] = { 0, 0, 0, 0 };
		int32 _egTL[4] = { 0, 0, 0, 0 };
		int32 _egTLv[4] = { 0, 0, 0, 0 };
	// Operator 0 ---
		int32& _lfoMod_0 = _lfoMod[0];
		int32& _lfoModSens_0 = _lfoModSens[0];
		int32& _lfoOutPM_0 = _lfoOutPM[0];
		int32& _lfoOutAM_0 = _lfoOutAM[0];
		int32*& _waveTable_0 = _waveTable[0];
		int32& _pgKey_0 = _pgKey[0];
		int32& _pgPhase_0 = _pgPhase[0];
		int32& _pgPhaseStep_0 = _pgPhaseStep[0];
		int32& _egState_0 = _egState[0];
		int32*& _egLevelTable_0 = _egLevelTable[0];
		int32& _egCount_0 = _egCount[0];
		int32& _egCountStep_0 = _egCountStep[0];
		int32& _egCountStepDR_0 = _egCountStepDR[0];
		int32& _egCountStepRR_0 = _egCountStepRR[0];
		int32& _egCntEnvSus_0 = _egCntEnvSus[0];
		int32& _egCntEnvBtm_0 = _egCntEnvBtm[0];
		int32& _egGate_0 = _egGate[0];
		int32& _egSL_0 = _egSL[0];
		int32& _egTL_0 = _egTL[0];
		int32& _egTLv_0 = _egTLv[0];
	// Operator 1 ---
		int32& _lfoMod_1 = _lfoMod[1];
		int32& _lfoModSens_1 = _lfoModSens[1];
		int32& _lfoOutPM_1 = _lfoOutPM[1];
		int32& _lfoOutAM_1 = _lfoOutAM[1];
		int32*& _waveTable_1 = _waveTable[1];
		int32& _pgKey_1 = _pgKey[1];
		int32& _pgPhase_1 = _pgPhase[1];
		int32& _pgPhaseStep_1 = _pgPhaseStep[1];
		int32& _egState_1 = _egState[1];
		int32*& _egLevelTable_1 = _egLevelTable[1];
		int32& _egCount_1 = _egCount[1];
		int32& _egCountStep_1 = _egCountStep[1];
		int32& _egCountStepDR_1 = _egCountStepDR[1];
		int32& _egCountStepRR_1 = _egCountStepRR[1];
		int32& _egCntEnvSus_1 = _egCntEnvSus[1];
		int32& _egCntEnvBtm_1 = _egCntEnvBtm[1];
		int32& _egGate_1 = _egGate[1];
		int32& _egSL_1 = _egSL[1];
		int32& _egTL_1 = _egTL[1];
		int32& _egTLv_1 = _egTLv[1];
	// Operator 2 ---
		int32& _lfoMod_2 = _lfoMod[2];
		int32& _lfoModSens_2 = _lfoModSens[2];
		int32& _lfoOutPM_2 = _lfoOutPM[2];
		int32& _lfoOutAM_2 = _lfoOutAM[2];
		int32*& _waveTable_2 = _waveTable[2];
		int32& _pgKey_2 = _pgKey[2];
		int32& _pgPhase_2 = _pgPhase[2];
		int32& _pgPhaseStep_2 = _pgPhaseStep[2];
		int32& _egState_2 = _egState[2];
		int32*& _egLevelTable_2 = _egLevelTable[2];
		int32& _egCount_2 = _egCount[2];
		int32& _egCountStep_2 = _egCountStep[2];
		int32& _egCountStepDR_2 = _egCountStepDR[2];
		int32& _egCountStepRR_2 = _egCountStepRR[2];
		int32& _egCntEnvSus_2 = _egCntEnvSus[2];
		int32& _egCntEnvBtm_2 = _egCntEnvBtm[2];
		int32& _egGate_2 = _egGate[2];
		int32& _egSL_2 = _egSL[2];
		int32& _egTL_2 = _egTL[2];
		int32& _egTLv_2 = _egTLv[2];
	// Operator 0 ---
		int32& _lfoMod_3 = _lfoMod[3];
		int32& _lfoModSens_3 = _lfoModSens[3];
		int32& _lfoOutPM_3 = _lfoOutPM[3];
		int32& _lfoOutAM_3 = _lfoOutAM[3];
		int32*& _waveTable_3 = _waveTable[3];
		int32& _pgKey_3 = _pgKey[3];
		int32& _pgPhase_3 = _pgPhase[3];
		int32& _pgPhaseStep_3 = _pgPhaseStep[3];
		int32& _egState_3 = _egState[3];
		int32*& _egLevelTable_3 = _egLevelTable[3];
		int32& _egCount_3 = _egCount[3];
		int32& _egCountStep_3 = _egCountStep[3];
		int32& _egCountStepDR_3 = _egCountStepDR[3];
		int32& _egCountStepRR_3 = _egCountStepRR[3];
		int32& _egCntEnvSus_3 = _egCntEnvSus[3];
		int32& _egCntEnvBtm_3 = _egCntEnvBtm[3];
		int32& _egGate_3 = _egGate[3];
		int32& _egSL_3 = _egSL[3];
		int32& _egTL_3 = _egTL[3];
		int32& _egTLv_3 = _egTLv[3];

	public:
		void keyOff() override;
		void soundOff() override;
	private:
		int32 _shiftEGState(int32);
	public:
		void keyOn(const int32&, const int32&, const int32&) override;
		void keyOnSampler(const int32&, const int32&, const int32&);
		void updPartBendSens(const int32&) override;
		void updPartPitchBend(const int32&) override;
		void updPartModulation(const int32&) override;
		void updateLfoFrequency(const int32&);
	public:
		void setup(SynthData*);
	// process ---
	protected:
		void _proc0(float* const&, float* const&, const int32&) override;
		void _proc1(float* const&, float* const&, const int32&) override;
		void _proc2(float* const&, float* const&, const int32&) override;
		void _proc3(float* const&, float* const&, const int32&) override;
		void _proc4(float* const&, float* const&, const int32&) override;
		void _proc5(float* const&, float* const&, const int32&) override;
		void _proc6(float* const&, float* const&, const int32&) override;
		void _proc7(float* const&, float* const&, const int32&) override;
		void _proc8(float* const&, float* const&, const int32&) override;
		void _proc9(float* const&, float* const&, const int32&) override;
		void _procA(float* const&, float* const&, const int32&) override;
		void _procB(float* const&, float* const&, const int32&) override;
		void _procC(float* const&, float* const&, const int32&) override;
		void _procD(float* const&, float* const&, const int32&) override;
		void _procE(float* const&, float* const&, const int32&) override;
		void _procF(float* const&, float* const&, const int32&) override;

		void _procSmpl1ch(float* const&, float* const&, const int32&) override;
		void _procSmpl2ch(float* const&, float* const&, const int32&) override;
	};

// Factory -----
private:	
	void _recreateVU(int32 maxPoly) override;
	void _resetKeyParams() override;
// Synthesizer - Data -----
private:
	SynthData* _prvSynthData = nullptr;
	SynthData* _synthData = nullptr;
	int32* waveTables = nullptr;  // 128 * 128
	SampleData* sampleDatas = nullptr;
public:
	TonePresetFM4* tonePresets = nullptr;  // [0 - 127](ProgramChange), [128 - 255](Percussion Key)
	SamplerPreset* samplerPresets = nullptr;
	
	void getWaveForm(int32 id, float* buffer, int32 bufferSize);
	void setWaveForm(int32 id, const float* buffer, int32 bufferSize);
	SampleData* getSampleData(int32 id){ return &sampleDatas[id]; }
	void setSampleData(int32 id, SampleData&& sampleData);
	void setSampleData(int32 id, const SampleData& sampleData, bool deepCopy = true);
	void loadPCMSamples(int32 id, float* wave, int32 samples, int32 ch);
	void loadPCMData(int32 id, uint8* dat, int32 sampleFrames, int32 ch, int32 bitRate, bool sign, bool bigEndian);
	void clearSampleData(int32 id);
	void factoryDataReset() override;
// Synthesizer - main -----
private:
	VoiceUnitFM4* _voiceUnitExt = nullptr;
// Constructor
public:
	SynthesizerFM4() : SynthesizerFM4(32, 440.0f, 1.0f, true){}
	SynthesizerFM4(int32 maxPoly = 32, float refPitch = 440.0f, float masVol = 1.0f, bool privateTableData = true);// Controls ---
// Copy Constructor
	SynthesizerFM4(const SynthesizerFM4&) = delete;
	SynthesizerFM4& operator = (const SynthesizerFM4&) = delete;
// Move Constructor
	SynthesizerFM4(SynthesizerFM4&&)noexcept;
	SynthesizerFM4& operator = (SynthesizerFM4&&)noexcept;
// Destructor
	~SynthesizerFM4();
// MidiEvent ---
private:	
	void _noteOn(const MidiEvent&) override;
	void _readSysExData(const MidiEvent&) override;
};
}}}
#endif